import os
import tqdm
import json
import random
import shutil
import argparse
from html_utils import HTML, save_gif
from subset_utils import get_absolute_path
from preresize_dataset import resize_dataset

# specify the root directory for data
parser = argparse.ArgumentParser(description="Process some inputs.")
parser.add_argument("--data_root", type=str, help="root folder to save data")
args = parser.parse_args()
data_root = args.data_root


def collect_trainset_from_json(json_path, save_root):
    with open(json_path, "r") as f:
        included_info = json.load(f)

    split = "train"

    # create dataset folder
    if not os.path.exists(save_root):
        os.makedirs(save_root)

    # delete the folder of split (if any) and create new folders
    if os.path.exists(os.path.join(save_root, split)):
        shutil.rmtree(os.path.join(save_root, split))

    os.makedirs(os.path.join(save_root, split))
    os.makedirs(os.path.join(save_root, split, "frameA"))
    os.makedirs(os.path.join(save_root, split, "frameB"))

    for fn, info in included_info.items():
        image_fn = "{}.png".format(fn)
        shutil.copy(
            get_absolute_path(info["paths"][0], data_root),
            os.path.join(save_root, split, "frameA", image_fn),
        )
        shutil.copy(
            get_absolute_path(info["paths"][1], data_root),
            os.path.join(save_root, split, "frameB", image_fn),
        )


def generate_trainset(info_path_dict, save_root, threshold_dict, mse_threshold):
    split = "train"

    # create dataset folder
    if not os.path.exists(save_root):
        os.makedirs(save_root)

    # delete the folder of split (if any) and create new folders
    if os.path.exists(os.path.join(save_root, split)):
        shutil.rmtree(os.path.join(save_root, split))

    os.makedirs(os.path.join(save_root, split))
    os.makedirs(os.path.join(save_root, split, "frameA"))
    os.makedirs(os.path.join(save_root, split, "frameB"))

    included_data = {}
    filtered_data = {}
    fn_dict = {}
    all_fn = []
    num_videos = 0

    for dataset, info_path in info_path_dict.items():
        fn_dict[dataset] = {}
        num = 0

        # read motion data from json file
        with open(info_path, "r") as f:
            flow_data = json.load(f)

        for idx, value in tqdm.tqdm(flow_data.items()):
            # try threshold for mse
            if value["mse"] <= mse_threshold:
                save = False
            else:
                # try percentile thresholds for motion
                for key, threshold in threshold_dict.items():
                    if value[key] > threshold:
                        save = False
                        break
                    else:
                        save = True

            if not save:
                # save info of filtered data
                filtered_data["{}_{}".format(dataset, str(idx))] = value

            else:
                # save info of included data and copy image files to new directory
                included_data["{}_{}".format(dataset, str(idx))] = value
                image_paths = value["paths"]
                image_fn = "{}_{}.png".format(dataset, str(num).zfill(5))
                fn_dict[dataset][idx] = {
                    "fn": image_fn,
                    "frameA": image_paths[0],
                    "frameB": image_paths[1],
                }
                all_fn.append(image_fn)
                shutil.copy(
                    get_absolute_path(image_paths[0], data_root),
                    os.path.join(save_root, split, "frameA", image_fn),
                )
                shutil.copy(
                    get_absolute_path(image_paths[1], data_root),
                    os.path.join(save_root, split, "frameB", image_fn),
                )
                num += 1

        print("num of videos from {}: {}".format(dataset, num))
        num_videos += num

    print("number of the selected videos: {}".format(num_videos))

    # write the info and filenames of filtered/included data into json files
    with open(os.path.join(save_root, "{}_files.json".format(split)), "w") as f:
        json.dump(fn_dict, f, indent=4)

    with open(os.path.join(save_root, "{}_fn.json".format(split)), "w") as f:
        json.dump(all_fn, f, indent=4)

    with open(os.path.join(save_root, "{}_filtered.json".format(split)), "w") as f:
        json.dump(filtered_data, f, indent=4)

    with open(os.path.join(save_root, "{}_included.json".format(split)), "w") as f:
        json.dump(included_data, f, indent=4)

    # visualize the gif of frame pairs of filtered/included data
    print("saving html to {}".format(os.path.join(save_root, "{}_vis".format(split))))
    html = HTML(os.path.join(save_root, "{}_vis".format(split)), "subset")
    html.add_header(save_root.split("/")[-1])

    num_vis = 30
    filtered_vis = random.sample(list(filtered_data.keys()), num_vis)
    included_vis = random.sample(list(included_data.keys()), num_vis)

    html.add_header(
        "split: {}, threshold: {}, {} (MSE)".format(
            split, str(threshold_dict), str(mse_threshold)
        )
    )
    html.add_header(
        "filtered videos, {}/{}".format(
            len(filtered_data), len(filtered_data) + len(included_data)
        )
    )
    ims = []
    txts = []
    links = []
    for n in filtered_vis:
        paths = filtered_data[n]["paths"]
        gif_path = save_gif(
            paths, os.path.join(save_root, "{}_vis".format(split), "images")
        )
        ims.append(gif_path)
        txts.append(str(filtered_data[n]))
        links.append(gif_path)
    html.add_images(ims, txts, links)

    html.add_header(
        "included videos, {}/{}".format(
            len(included_data), len(filtered_data) + len(included_data)
        )
    )
    ims = []
    txts = []
    links = []
    for n in included_vis:
        paths = included_data[n]["paths"]
        gif_path = save_gif(
            paths, os.path.join(save_root, "{}_vis".format(split), "images")
        )
        ims.append(gif_path)
        txts.append(str(included_data[n]))
        links.append(gif_path)
    html.add_images(ims, txts, links)
    html.save()


def generate_testset(
    info_path_dict, save_root, threshold_dict, mse_threshold, num_per_dataset=100
):
    split = "test"

    # create dataset folder
    if not os.path.exists(save_root):
        os.makedirs(save_root)

    # delete the folder of split (if any) and create new folders
    if os.path.exists(os.path.join(save_root, split)):
        shutil.rmtree(os.path.join(save_root, split))

    os.makedirs(os.path.join(save_root, split))
    os.makedirs(os.path.join(save_root, split, "frameA"))
    os.makedirs(os.path.join(save_root, split, "frameB"))

    included_data = {}
    filtered_data = {}
    fn_dict = {}
    all_fn = []

    for dataset, info_path in info_path_dict.items():
        if dataset in ["tao_n1", "tao_n5"]:
            num_per_dataset_ = int(num_per_dataset / 2)

            # collect from every subsets for TAO dataset
            temp_filtered_data = {
                "ArgoVerse": {},
                "BDD": {},
                "Charades": {},
                "LaSOT": {},
                "YFCC100M": {},
            }
            temp_included_data = {
                "ArgoVerse": {},
                "BDD": {},
                "Charades": {},
                "LaSOT": {},
                "YFCC100M": {},
            }

            with open(info_path, "r") as f:
                flow_data = json.load(f)

            for idx, value in tqdm.tqdm(flow_data.items()):
                image_paths = value["paths"]
                sub_name = image_paths[0].split("/")[-3]

                # try threshold for mse
                if value["mse"] <= mse_threshold:
                    save = False
                else:
                    # try percentile thresholds for motion
                    for key, threshold in threshold_dict.items():
                        if value[key] > threshold:
                            save = False
                            break
                        else:
                            save = True

                if not save:
                    # save info of filtered data
                    temp_filtered_data[sub_name][idx] = value
                    filtered_data["{}_{}".format(dataset, str(idx))] = value

                else:
                    # save info of included data
                    temp_included_data[sub_name][idx] = value
                    included_data["{}_{}".format(dataset, str(idx))] = value

            for key, value in temp_included_data.items():
                # randomly select from included data, save info and copy images to new folder
                fn_dict[dataset + "_" + key] = {}
                all_included = list(value.keys())
                random.shuffle(all_included)
                for i, idx in enumerate(all_included[:num_per_dataset_]):
                    image_paths = value[idx]["paths"]
                    image_fn = "{}_{}_{}.png".format(dataset, key, str(i).zfill(5))
                    fn_dict[dataset + "_" + key][i] = {
                        "fn": image_fn,
                        "frameA": image_paths[0],
                        "frameB": image_paths[1],
                    }
                    all_fn.append(image_fn)
                    shutil.copy(
                        get_absolute_path(image_paths[0], data_root),
                        os.path.join(save_root, split, "frameA", image_fn),
                    )
                    shutil.copy(
                        get_absolute_path(image_paths[1], data_root),
                        os.path.join(save_root, split, "frameB", image_fn),
                    )

        elif dataset in ["youtube_vos", "vimeo", "davis", "uvo_n1", "uvo_n5"]:
            num_per_dataset_ = (
                int(num_per_dataset / 2) if "uvo" in dataset else num_per_dataset
            )
            fn_dict[dataset] = {}

            with open(info_path, "r") as f:
                flow_data = json.load(f)
            for idx, value in tqdm.tqdm(flow_data.items()):
                # try threshold for mse
                if value["mse"] <= mse_threshold:
                    save = False
                else:
                    # try percentile thresholds for motion
                    for key, threshold in threshold_dict.items():
                        if value[key] > threshold:
                            save = False
                            break
                        else:
                            save = True
                if not save:
                    # save info of filtered data
                    filtered_data["{}_{}".format(dataset, str(idx))] = value

                else:
                    # save info of included data
                    included_data["{}_{}".format(dataset, str(idx))] = value

            # randomly select from included data, save info and copy images to new folder
            all_included = list(included_data.keys())
            random.shuffle(all_included)
            print(dataset)
            for i, idx in enumerate(all_included[:num_per_dataset_]):
                image_paths = included_data[idx]["paths"]
                image_fn = "{}_{}.png".format(dataset, str(i).zfill(5))
                fn_dict[dataset][i] = {
                    "fn": image_fn,
                    "frameA": image_paths[0],
                    "frameB": image_paths[1],
                }
                all_fn.append(image_fn)
                shutil.copy(
                    get_absolute_path(image_paths[0], data_root),
                    os.path.join(save_root, split, "frameA", image_fn),
                )
                shutil.copy(
                    get_absolute_path(image_paths[1], data_root),
                    os.path.join(save_root, split, "frameB", image_fn),
                )

    print(len(all_fn))

    # save info and filenames
    with open(os.path.join(save_root, "{}_files.json".format(split)), "w") as f:
        json.dump(fn_dict, f, indent=4)

    with open(os.path.join(save_root, "{}_fn.json".format(split)), "w") as f:
        json.dump(all_fn, f, indent=4)

    with open(os.path.join(save_root, "{}_filtered.json".format(split)), "w") as f:
        json.dump(filtered_data, f, indent=4)

    with open(os.path.join(save_root, "{}_included.json".format(split)), "w") as f:
        json.dump(included_data, f, indent=4)

    # visualize with gif of frame pairs of filtered/included data
    print("saving html to {}".format(os.path.join(save_root, "{}_vis".format(split))))
    html = HTML(os.path.join(save_root, "{}_vis".format(split)), "subset")
    html.add_header(save_root.split("/")[-1])

    num_vis = 10
    filtered_vis = random.sample(list(filtered_data.keys()), num_vis)
    included_vis = random.sample(list(included_data.keys())[:num_per_dataset], num_vis)

    html.add_header(
        "split: {}, threshold: {}, {} (MSE)".format(
            split, str(threshold_dict), str(mse_threshold)
        )
    )
    html.add_header(
        "filtered videos, {}/{}".format(
            len(filtered_data), len(filtered_data) + len(included_data)
        )
    )
    ims = []
    txts = []
    links = []
    for n in filtered_vis:
        paths = filtered_data[n]["paths"]
        gif_path = save_gif(
            paths, os.path.join(save_root, "{}_vis".format(split), "images")
        )
        ims.append(gif_path)
        txts.append(str(filtered_data[n]))
        links.append(gif_path)
    html.add_images(ims, txts, links)

    html.add_header(
        "included videos, {}/{}".format(
            len(included_data), len(filtered_data) + len(included_data)
        )
    )
    ims = []
    txts = []
    links = []
    for n in included_vis:
        paths = included_data[n]["paths"]
        gif_path = save_gif(
            paths, os.path.join(save_root, "{}_vis".format(split), "images")
        )
        ims.append(gif_path)
        txts.append(str(included_data[n]))
        links.append(gif_path)
    html.add_images(ims, txts, links)
    html.save()


if __name__ == "__main__":
    dataset_path = os.path.join(data_root, "flowmag_raw")

    # thresholds of motion and mse
    threshold_dict = {"max_dist": 20, "80th": 2, "0.01st": 0.1}
    mse_threshold = 10

    # path to the files of the motion and mse
    info_path_dict = {
        "youtube_vos": os.path.join(data_root, "flow_data/youtube_vos_train.json"),
        "davis": os.path.join(data_root, "flow_data/davis_trainval.json"),
        "vimeo": os.path.join(data_root, "flow_data/vimeo_train.json"),
        "tao_n1": os.path.join(data_root, "flow_data/tao_train_n1.json"),
        "tao_n5": os.path.join(data_root, "flow_data/tao_train_n5.json"),
        "uvo_n1": os.path.join(data_root, "flow_data/uvo_train_n1.json"),
        "uvo_n5": os.path.join(data_root, "flow_data/uvo_train_n5.json"),
    }
    # collect train set according to the given threshold
    generate_trainset(info_path_dict, dataset_path, threshold_dict, mse_threshold)

    info_path_dict = {
        "youtube_vos": os.path.join(data_root, "flow_data/youtube_vos_test.json"),
        "davis": os.path.join(data_root, "flow_data/davis_test.json"),
        "vimeo": os.path.join(data_root, "flow_data/vimeo_test.json"),
        "tao_n1": os.path.join(data_root, "flow_data/tao_val_n1.json"),
        "tao_n5": os.path.join(data_root, "flow_data/tao_val_n5.json"),
        "uvo_n1": os.path.join(data_root, "flow_data/uvo_test_n1.json"),
        "uvo_n5": os.path.join(data_root, "flow_data/uvo_test_n5.json"),
    }
    # collect test set according to the given threshold, 100 samples for every subset
    generate_testset(
        info_path_dict, dataset_path, threshold_dict, mse_threshold, num_per_dataset=100
    )

    # resize the frame pairs to get final dataset
    final_path = os.path.join(data_root, "flowmag_data")
    resize_dataset(dataset_path, "train", final_path)
    resize_dataset(dataset_path, "test", final_path)



import torch
import torch.nn as nn
from einops import rearrange, repeat

from neurovc.contrib.flowmag.flow_utils import GMFlow, PWC, RAFT


class Metrics(nn.Module):
    def __init__(self, config):
        super().__init__()

        # Choose flow model to calculate metrics
        if config.train.flow_model == "raft":
            print("using RAFT")
            self.flow_net = RAFT(
                model=config.train.flow_model_type, num_iters=config.train.raft_iters
            )
        elif config.train.flow_model == "gmflow":
            print("using GMFlow")
            self.flow_net = GMFlow(model=config.train.flow_model_type)
        elif config.train.flow_model == "pwcnet":
            print("using PWCNet")
            self.flow_net = PWC()
        else:
            raise NotImplementedError

        # Use L1 for metrics calculation
        self.criterion = nn.L1Loss()

    def compute_flow(self, im, frames, backward=False):
        """
        Given an image (b, c, h, w) and frames (b, c, t, h, w) we compute
            the flow between the image and all frames and return them
            in the shape (b, t, 2, h, w)
        If backward == True, then reverse the direction: compute flow from
            all frames to single image
        """
        b, c, t, h, w = frames.shape

        # Repeat and rearrange as needed
        im = repeat(im, "b c h w -> (b num_frames) c h w", num_frames=t)
        frames = rearrange(frames, "b c t h w -> (b t) c h w")

        # Compute flow
        if backward:
            flow = self.flow_net(frames, im)
        else:
            flow = self.flow_net(im, frames)

        # Rearrange to separate batch and time dim
        flow = rearrange(flow, "(b t) d h w -> b t d h w", b=b, t=t)

        return flow

    def compute_metrics(self, flow_pred, flow_src, alphas):
        # Calculate motion error in the way of mag_loss
        motion_error = self.criterion(flow_pred, alphas * flow_src)

        # Calculate motion error (ratio) between flow magnitudes ratio and alpha
        dist_pred = torch.sqrt(flow_pred[:, :, 0] ** 2 + flow_pred[:, :, 1] ** 2)
        dist_src = torch.sqrt(flow_src[:, :, 0] ** 2 + flow_src[:, :, 1] ** 2)
        mag_error = self.criterion(dist_pred / (dist_src + 0.0001), alphas)

        return motion_error, mag_error

    def forward(self, pred, src, alphas):
        """
        pred: magnified future frame predictions (b, c, t-1, h, w)
        src: original frames (b, c, t, h, w)
        alphas: magnification factors (per batch)
        """

        # Add dims to alpha to match flow (b, t, d, h, w)
        if len(alphas.shape) == 1:
            # Only batch dim, constant alpha per video
            alphas = alphas.view(-1, 1, 1, 1, 1)
        elif len(alphas.shape) == 3:
            # Spatially varying alpha, shape (b, h, w)
            b, h, w = alphas.shape
            alphas = alphas.view(b, 1, 1, h, w)

        # Get flows
        anchor_frame = src[:, :, 0]
        flow_pred = self.compute_flow(anchor_frame, pred)
        flow_src = self.compute_flow(anchor_frame, src[:, :, 1:])

        # Compute metrics
        motion_error, mag_error = self.compute_metrics(flow_pred, flow_src, alphas)

        return motion_error, mag_error



import torch
from torch.nn.modules.module import Module
from torch.autograd import Function
import correlation_cuda


class CorrelationFunction(Function):
    def __init__(
        self,
        pad_size=3,
        kernel_size=3,
        max_displacement=20,
        stride1=1,
        stride2=2,
        corr_multiply=1,
    ):
        super(CorrelationFunction, self).__init__()
        self.pad_size = pad_size
        self.kernel_size = kernel_size
        self.max_displacement = max_displacement
        self.stride1 = stride1
        self.stride2 = stride2
        self.corr_multiply = corr_multiply
        # self.out_channel = ((max_displacement/stride2)*2 + 1) * ((max_displacement/stride2)*2 + 1)

    def forward(self, input1, input2):
        self.save_for_backward(input1, input2)

        with torch.cuda.device_of(input1):
            rbot1 = input1.new()
            rbot2 = input2.new()
            output = input1.new()

            correlation_cuda.forward(
                input1,
                input2,
                rbot1,
                rbot2,
                output,
                self.pad_size,
                self.kernel_size,
                self.max_displacement,
                self.stride1,
                self.stride2,
                self.corr_multiply,
            )

        return output

    def backward(self, grad_output):
        input1, input2 = self.saved_tensors

        with torch.cuda.device_of(input1):
            rbot1 = input1.new()
            rbot2 = input2.new()

            grad_input1 = input1.new()
            grad_input2 = input2.new()

            correlation_cuda.backward(
                input1,
                input2,
                rbot1,
                rbot2,
                grad_output,
                grad_input1,
                grad_input2,
                self.pad_size,
                self.kernel_size,
                self.max_displacement,
                self.stride1,
                self.stride2,
                self.corr_multiply,
            )

        return grad_input1, grad_input2


class Correlation(Module):
    def __init__(
        self,
        pad_size=0,
        kernel_size=0,
        max_displacement=0,
        stride1=1,
        stride2=2,
        corr_multiply=1,
    ):
        super(Correlation, self).__init__()
        self.pad_size = pad_size
        self.kernel_size = kernel_size
        self.max_displacement = max_displacement
        self.stride1 = stride1
        self.stride2 = stride2
        self.corr_multiply = corr_multiply

    def forward(self, input1, input2):
        result = CorrelationFunction(
            self.pad_size,
            self.kernel_size,
            self.max_displacement,
            self.stride1,
            self.stride2,
            self.corr_multiply,
        )(input1, input2)

        return result



import torch
import torch.nn as nn
import torch.nn.functional as F


# Crecit: https://github.com/simonmeister/UnFlow/blob/master/src/e2eflow/core/losses.py
def TernaryLoss(im, im_warp, max_distance=1):
    patch_size = 2 * max_distance + 1

    def _rgb_to_grayscale(image):
        grayscale = (
            image[:, 0, :, :] * 0.2989
            + image[:, 1, :, :] * 0.5870
            + image[:, 2, :, :] * 0.1140
        )
        return grayscale.unsqueeze(1)

    def _ternary_transform(image):
        intensities = _rgb_to_grayscale(image) * 255
        out_channels = patch_size * patch_size
        w = torch.eye(out_channels).view((out_channels, 1, patch_size, patch_size))
        weights = w.type_as(im)
        patches = F.conv2d(intensities, weights, padding=max_distance)
        transf = patches - intensities
        transf_norm = transf / torch.sqrt(0.81 + torch.pow(transf, 2))
        return transf_norm

    def _hamming_distance(t1, t2):
        dist = torch.pow(t1 - t2, 2)
        dist_norm = dist / (0.1 + dist)
        dist_mean = torch.mean(dist_norm, 1, keepdim=True)  # instead of sum
        return dist_mean

    def _valid_mask(t, padding):
        n, _, h, w = t.size()
        inner = torch.ones(n, 1, h - 2 * padding, w - 2 * padding).type_as(t)
        mask = F.pad(inner, [padding] * 4)
        return mask

    t1 = _ternary_transform(im)
    t2 = _ternary_transform(im_warp)
    dist = _hamming_distance(t1, t2)
    mask = _valid_mask(im, max_distance)

    return dist * mask


def SSIM(x, y, md=1):
    patch_size = 2 * md + 1
    C1 = 0.01**2
    C2 = 0.03**2

    mu_x = nn.AvgPool2d(patch_size, 1, 0)(x)
    mu_y = nn.AvgPool2d(patch_size, 1, 0)(y)
    mu_x_mu_y = mu_x * mu_y
    mu_x_sq = mu_x.pow(2)
    mu_y_sq = mu_y.pow(2)

    sigma_x = nn.AvgPool2d(patch_size, 1, 0)(x * x) - mu_x_sq
    sigma_y = nn.AvgPool2d(patch_size, 1, 0)(y * y) - mu_y_sq
    sigma_xy = nn.AvgPool2d(patch_size, 1, 0)(x * y) - mu_x_mu_y

    SSIM_n = (2 * mu_x_mu_y + C1) * (2 * sigma_xy + C2)
    SSIM_d = (mu_x_sq + mu_y_sq + C1) * (sigma_x + sigma_y + C2)
    SSIM = SSIM_n / SSIM_d
    dist = torch.clamp((1 - SSIM) / 2, 0, 1)
    return dist


def gradient(data):
    D_dy = data[:, :, 1:] - data[:, :, :-1]
    D_dx = data[:, :, :, 1:] - data[:, :, :, :-1]
    return D_dx, D_dy


def smooth_grad_1st(flo, image, alpha):
    img_dx, img_dy = gradient(image)
    weights_x = torch.exp(-torch.mean(torch.abs(img_dx), 1, keepdim=True) * alpha)
    weights_y = torch.exp(-torch.mean(torch.abs(img_dy), 1, keepdim=True) * alpha)

    dx, dy = gradient(flo)

    loss_x = weights_x * dx.abs() / 2.0
    loss_y = weights_y * dy.abs() / 2

    return loss_x.mean() / 2.0 + loss_y.mean() / 2.0


def smooth_grad_2nd(flo, image, alpha):
    img_dx, img_dy = gradient(image)
    weights_x = torch.exp(-torch.mean(torch.abs(img_dx), 1, keepdim=True) * alpha)
    weights_y = torch.exp(-torch.mean(torch.abs(img_dy), 1, keepdim=True) * alpha)

    dx, dy = gradient(flo)
    dx2, dxdy = gradient(dx)
    dydx, dy2 = gradient(dy)

    loss_x = weights_x[:, :, :, 1:] * dx2.abs()
    loss_y = weights_y[:, :, 1:, :] * dy2.abs()

    return loss_x.mean() / 2.0 + loss_y.mean() / 2.0



# todo: rewrite
# from IPython import parallel
import ipyparallel as parallel
import util as ut
import os
import sys
import time

# from IPython.parallel.util import interactive
import subprocess
import numpy as np
import imp

try:
    from IPython.parallel.util import interactive
except Exception:
    try:
        from ipyparallel.util import interactive
    except Exception:

        def interactive(func):
            return func


ip_parallel = parallel

p = None
part_info = None
builtin_map = __builtins__["map"]


def relative_module(m):
    return hasattr(m, "__file__") and (
        (not m.__file__.startswith("/")) or m.__file__.startswith(os.getcwd())
    )


def safe_reload(m):
    if relative_module(m) and (not hasattr(m, "__no_autoreload__")):
        print("reloading")
        imp.reload(m)


def run_reload():
    excludes = set(["__main__", "autoreload"])
    for name, m in list(sys.modules.items()):
        if (
            m
            and relative_module(m)
            and (name not in excludes)
            and (not hasattr(m, "__no_autoreload__"))
        ):
            imp.reload(m)


class Parallel:
    def __init__(self, config=None, rc=None, view=None, bview=None):
        self.config = config
        self._rc = rc
        self._view = view
        self._bview = bview

    def init_rc(self, reset=False):
        if reset or (self._rc is None):
            if self.config is None:
                self._rc = ip_parallel.Client()
            else:
                self._rc = ip_parallel.Client(self.config)
            self._view = self._rc.direct_view()
            # self._view = self._rc.load_balanced_view()
            self._bview = self._rc.load_balanced_view()

    def view(self):
        self.init_rc()
        return self._view

    def bview(self):
        self.init_rc()
        return self._bview

    def rc(self):
        self.init_rc()
        return self._rc

    def per_machine_subset(self, n_per_machine):
        targets = self.targets_by_hostname()
        target_subset = ut.flatten(
            [ut.sample_at_most(ts, n_per_machine) for ts in list(targets.values())]
        )
        return Parallel(self.config, self.rc(), self.rc()[target_subset])

    def partition(self, n, i):
        target_subset = [j for j in self.rc()[:].targets if j % n == i]
        return Parallel(self.config, self.rc(), self.rc()[target_subset])

    def targets_by_hostname(self):
        # initialize parallel_rc
        self.import_modules(["subprocess"], False)
        targets_by_hostname = {}
        for t in self.rc()[:].targets:
            v = self.rc()[[t]]
            [hostname] = v.map_sync(
                interactive(
                    lambda x: subprocess.Popen(
                        "hostname", shell=True, stdout=subprocess.PIPE
                    )
                    .stdout.read()
                    .rstrip()
                ),
                [1],
            )
            if hostname not in targets_by_hostname:
                targets_by_hostname[hostname] = []
            targets_by_hostname[hostname].append(t)
        return targets_by_hostname

    def execute(self, code):
        r = self.view().execute(code)
        r.wait()
        if not r.successful():
            print("iputil.execute error:")
            print(r.result())
        assert r.successful()

    # def list_machines(self):
    #   self.view().execute('import subprocess, os').wait()
    #   res = self.view().map_sync(interactive(lambda x : subprocess.Popen('hostname', shell=True,
    #                                                                 stdout=subprocess.PIPE).stdout.read().rstrip()), range(200))
    #   print 'Running on:'
    #   ut.prn_lines(set(res))

    def list_machines(self):
        self.view().execute("import subprocess, os").wait()
        # res = self.view().map_sync(interactive(lambda x : (subprocess.Popen('hostname', shell=True, stdout=subprocess.PIPE).stdout.read().rstrip(),
        #                                                    os.getpid())), range(2000))
        res = self.view().map_sync(
            lambda x: (
                subprocess.Popen("hostname", shell=True, stdout=subprocess.PIPE)
                .stdout.read()
                .rstrip(),
                os.getpid(),
            ),
            list(range(2000)),
        )
        print("Running on:")
        machines = {}
        for x, p in res:
            if x not in machines:
                machines[x] = set()
            machines[x].add(p)
        for x, ps in list(machines.items()):
            print(x, len(ps))

    def send_var(self, var_name, val):
        res = self.view().push({var_name: val})
        res.wait()
        assert res.successful()

    def import_modules(self, packages, reload=False):
        packages = [("aolib.parutil", "parutil")] + list(packages)
        import_cmd = ""
        reload_cmd = ""
        for p in packages:
            if len(import_cmd) != 0:
                import_cmd += ", "
                reload_cmd += "; "
            if isinstance(p, tuple):
                import_cmd += "%s as %s" % p
                reload_cmd += "parutil.safe_reload(%s)" % p[1]
            else:
                import_cmd += p
                reload_cmd += "parutil.safe_reload(%s)" % p

        if reload:
            cmd = "import %s; %s" % (import_cmd, reload_cmd)
        else:
            cmd = "import %s" % import_cmd
        self.execute(cmd)

    def reload_modules(self):
        self.execute(
            "from aolib import parutil as parutil, areload; parutil.run_reload()"
        )
        # self.execute('import aolib.parutil as parutil, areload; parutil.run_reload()')

    def send_var_fs(self, var_name, val):
        fname = ut.make_temp_nfs(".pk")
        ut.save(fname, val)
        res = self.view().execute(
            'import util as ut; ut.wait_for_file("%s", 30); %s = ut.load("%s")'
            % (fname, var_name, fname)
        )
        res.wait()
        assert res.successful()
        os.remove(fname)

    def clear(self):
        self.view().results.clear()
        self.rc().results.clear()
        self.rc().metadata.clear()

    # def maybe_map_sync(self, parallel, f, xs, imports = []):
    #   if parallel:
    #     return self.map_sync(f, xs, imports)
    #   else:
    #     # assume imports have already been loaded if not parallel
    #     if type(xs) == type((1,)):
    #       return map(f, *xs)
    #     else:
    #       return map(f, xs)

    # def maybe_map_sync(self, parallel, f, xs, imports = [], vars = {}):
    #   if parallel:
    #     return self.map_sync(f, xs, imports)
    #   else:
    #     # todo: is it better to save these temporarily? or would it only make it more confusing?
    #     f.func_globals.update(vars)
    #     # assume imports have already been loaded if not parallel
    #     if type(xs) == type((1,)):
    #       return map(f, *xs)
    #     else:
    #       return map(f, xs)

    def maybe_map_sync(self, parallel, f, xs, vars={}, imports=[], do_reload=True):
        if parallel:
            return self.map_sync(f, xs, vars, imports, do_reload=do_reload)
        else:
            # todo: is it better to save these temporarily? or would it only make it more confusing?
            if len(vars):
                f.__globals__.update(vars)
            # assume imports have already been loaded if not parallel
        if isinstance(xs, tuple):
            return builtin_map(f, *xs)
        else:
            return builtin_map(f, xs)

    def map(self, parallel, f, xs, vars={}, imports=[], do_reload=True):
        return self.maybe_map_sync(parallel, f, xs, vars, imports, do_reload=do_reload)

    def amap(self, *args, **kwargs):
        return np.array(self.map(*args, **kwargs))

    def map_sync(
        self,
        f,
        xs,
        vars={},
        imports=[],
        do_reload=True,
        do_reset=False,
        use_bview=False,
    ):
        # if do_reset:
        #   self.init_rc(do_reset)

        if len(imports):
            self.import_modules(imports, True)
        if do_reload:
            self.reload_modules()

        if len(xs) == 0:
            return []

        if use_bview:
            view = self.view() if do_reload else self.bview()
        else:
            view = self.view()
        # interpret tuple as multiple args
        if isinstance(xs, tuple):
            res = view.map_sync(f, *xs)
        else:
            res = view.map_sync(f, xs)

        self.clear()

        return res

    def abort(self):
        self.view().abort()

    #  def test(self, do = True, max_wait = 30):
    def test(self, do=True, max_wait=60):
        if do:
            a = self.view().map_async(lambda x: x, list(range(100)))
            start = time.time()
            while time.time() - start <= max_wait:
                if a.ready():
                    return True
                else:
                    time.sleep(1)
            raise RuntimeError("test() timed out after %s seconds!" % max_wait)


__no_autoreload__ = True


def init(do=True, local=True):
    if do:
        global p, part_info
        if local:
            config_file = None
        else:
            config_file = "/data/vision/billf/aho-billf/ipcluster_profile/security/ipcontroller-client.json"
        if p is not None and p._rc is not None:
            p._rc.close()
        p = Parallel(config_file)
        if part_info is not None:
            print("parutil partition:", part_info)
            p = p.partition(*part_info)
        for f in dir(p):
            if not f.startswith("_") and hasattr(getattr(p, f), "__call__"):
                globals()[f] = getattr(p, f)


init()

# def restart():
#   os.system('./cluster.sh')
#   init()


def global_partition(n=None, i=None):
    global p, part_info
    if n is None:
        part_info = None
    else:
        part_info = (n, i)
    init()


def test_partition():
    def f(x):
        from neurovc.contrib.flowmag.flow_models.pwcnet.aolib import parutil

        return sum(parutil.partition(2, 1).map_sync(lambda x: x, list(range(10))))

    from neurovc.contrib.flowmag.flow_models.pwcnet.aolib import parutil

    parutil.map_sync(lambda x: x, list(range(10000)))
    assert sum(parutil.partition(2, 0).map_sync(f, list(range(1000)))) == 1000 * sum(
        range(10)
    )


def reset(parallel=True):
    init(parallel)
    if p is not None:
        p.test(do=parallel)



import wave
import aolib.util as ut
import os
import numpy as np
import aolib.img as ig
import aolib.imtable as imtable
import pylab as pl
import copy
import scipy.io.wavfile


class converted_wav:
    def __init__(self, in_fname):
        self.in_fname = in_fname
        self.out_fname = ut.make_temp(".wav")
        os.remove(self.out_fname)

    def __enter__(self):
        # if 0 != os.system('ffmpeg -loglevel warning -i "%s" -acodec pcm_s16le "%s"' % (self.in_fname, self.out_fname)):
        if 0 != os.system(
            'ffmpeg -loglevel warning -i "%s" -acodec pcm_s32le "%s"'
            % (self.in_fname, self.out_fname)
        ):
            raise RuntimeError("Could not convert wav file")
        return self.out_fname

    def __exit__(self, type, value, traceback):
        assert self.in_fname != self.out_fname
        os.remove(self.out_fname)


def load_wav(wav_fname):
    # deprecated
    with converted_wav(wav_fname) as fname:
        return scipy.io.wavfile.read(fname)


def load_sound(wav_fname):
    rate, samples = scipy.io.wavfile.read(wav_fname)
    times = (1.0 / rate) * np.arange(len(samples))
    return Sound(times, rate, samples)


def load_sound_uint8(wav_fname):
    rate, samples = scipy.io.wavfile.read(wav_fname)
    assert samples.dtype == np.uint8
    samples = np.array(samples, "float32") / 128.0 - 0.5
    times = (1.0 / rate) * np.arange(len(samples))
    return Sound(times, rate, samples)


def time_idx_interval(times, t1, t2):
    ti1 = np.searchsorted(times, t1)
    ti2 = np.searchsorted(times, t2)
    return ti1, ti2


def db(x):
    return 20 * np.log10(x)


def vis_specgram(
    freqs,
    sft,
    times,
    lo_db=-90.0,
    hi_db=-10.0,
    fig_=None,
    freq_lo_hz=1.0,
    freq_hi_hz=20000.0,
    width_in=4.0,
    time_lo=None,
    time_hi=None,
    plot_fn=(lambda fig, ax: None),
):
    # fig = pl.figure(frameon=False) if fig_ is None else fig_
    fig = pl.gcf() if fig_ is None else fig_
    fig.clf()
    fig.set_facecolor((1, 1, 1, 1))
    ax = fig.gca()

    # f1, f2 = np.min(freqs), np.max(freqs)
    # f1 = freq_lo_hz
    # f2 = np.max(freqs) if freq_hi_hz is None else freq_hi_hz

    # t1 = times[0] if time_lo is None else time_lo
    # t2 = times[-1] if time_hi is None else time_hi
    t1 = times[0]
    t2 = times[-1]

    sft = sft.T
    # lsp = (sft[(f1 <= freqs) & (freqs <= f2)][:, (t1 <= times) & (times <= t2)]).copy()
    # fi1 = np.searchsorted(freqs, f1)
    # fi2 = np.searchsorted(freqs, f2)
    ti1 = np.searchsorted(times, t1)
    ti2 = np.searchsorted(times, t2)
    # asdf
    # lsp = (sft[fi1 : fi2][:, ti1 : ti2]).copy()
    lsp = (sft[:, ti1:ti2]).copy()
    ok = lsp > 0
    lsp[ok] = db(lsp[ok])
    lsp[-ok] = db(0.0000000001)

    # vis_ok = ut.clip_rescale(lsp, lo_db, hi_db)
    vis_ok = lsp
    # print np.min(lsp), np.max(lsp)
    # ax.imshow(vis_ok, vmin = 0., vmax = 1., cmap = pl.cm.jet, extent = (t1, t2, f1, f2), aspect = 'auto')
    # ax.imshow(vis_ok, vmin = 0., vmax = 1., cmap = pl.cm.gray_r, extent = (t1, t2, f1, f2), aspect = 'auto')
    # ax.imshow(vis_ok, vmin = lo_db, vmax = hi_db, cmap = pl.cm.gray_r,
    #           extent = (t1, t2, f1, f2), aspect = 'auto')
    ax.imshow(
        vis_ok,
        vmin=lo_db,
        vmax=hi_db,
        cmap=pl.cm.gray_r,
        extent=(t1, t2, freqs[-1], freqs[0]),
        aspect="auto",
    )
    # ax.set_yscale('log')
    ax.set_yscale("linear")
    ax.set_xscale("linear")
    ax.set_axis_bgcolor((1, 1, 1, 1))
    ax.set_ylim([0, 20000])
    ax.set_xlim([time_lo, time_hi])

    plot_fn(fig, ax)

    fig.tight_layout(pad=0)
    # ret = ig.from_fig(ax)

    ret = ig.from_fig(fig)
    # if fig_ is None:
    #   pl.close(fig)

    # w, h = fig.get_size_inches()
    # fig.set_size_inches(float(width_in), float(width_in)/w * h)

    return ret[:, :, :3]  # ret[:, :, :3]


def specgram_max():
    sft_hi = 0.1
    # This number was computed roughly using the following procedure:
    # snd = scan.sound().normalized()
    # specgram = sound.make_specgram(snd.samples, snd.rate)
    # np.percentile(specgram[1], 99.5)
    return sft_hi


def centered_sound(sound, ts, duration=1.5):
    tidx1, tidx2 = time_idx_interval(sound.times, ts - duration / 2, ts + duration / 2)
    return sound[tidx1:tidx2].normalized()


# def vis_specgram_at_time(sound, ts, duration = 0.5):
#   sound_sub = centered_sound(sound, ts, duration)
#   spec = make_specgram(sound_sub.samples, sound_sub.rate, sample_times = sound_sub.times, noverlap = 2000)
#   def plot_fn(fig, ax, ts = ts):
#     ax.plot([ts, ts], ax.get_ylim(), alpha = 0.25, color = 'w', lw = 2)

#   return vis_specgram(spec[0], spec[1], spec[2], hi = specgram_max(), plot_fn = plot_fn)

# def vis_specgram_at_times(snd, times, window_sec, par = 0):
#   snd = snd.normalized().to_mono()
#   freqs, sft, sft_times = make_specgram(snd.samples, snd.rate, sample_times = snd.times, noverlap = 2000)
#   ims = []
#   for t in times:
#     def plot_fn(fig, ax, t = t):
#       ax.plot([t, t], ax.get_ylim(), alpha = 0.25, color = 'blue', lw = 2)
#     time_lo = t - window_sec/2.
#     time_hi = t + window_sec/2.
#     ims.append(vis_specgram(freqs, sft, sft_times, time_lo = time_lo, time_hi = time_hi, plot_fn = plot_fn))
#   return ims


def vis_specgram_at_times(
    snd, times, window_sec, par=0, im_shape=(500, 888), compress_ims=False
):
    snd = snd.normalized().to_mono()
    spec = make_specgram(snd.samples, snd.rate, sample_times=snd.times, noverlap=2000)

    def f(t, spec=spec, window_sec=window_sec, im_shape=im_shape):
        freqs, sft, sft_times = spec

        def plot_fn(fig, ax, t=t):
            ax.plot([t, t], ax.get_ylim(), alpha=0.25, color="blue", lw=2)

        time_lo = t - window_sec / 2.0
        time_hi = t + window_sec / 2.0
        vsc = vis_specgram(
            freqs, sft, sft_times, time_lo=time_lo, time_hi=time_hi, plot_fn=plot_fn
        )
        return ig.compress(ig.scale(vsc, im_shape))

    if not compress_ims:
        mapper = par.map if hasattr(par, "map") else map
        ims = [ig.uncompress(res) for res in mapper(f, times)]
    else:
        ims = list(map(f, times))
    return ims


# def make_mono(x):
#   if np.ndim(x) > 1:
#     assert np.ndim(x) == 2
#     return np.sum(x, axis = 1, dtype = 'double')
#   else:
#     return np.array(x, dtype = 'double')


def make_mono(x):
    if np.ndim(x) > 1:
        assert np.ndim(x) == 2
        return np.mean(x, axis=1, dtype="double")
    else:
        return np.array(x, dtype="double")


def stfft(x, nfft, noverlap, win, par=0):
    x = make_mono(x)
    step = nfft - noverlap
    n = win.shape[0]
    res = []
    win_starts = np.arange(0, x.shape[0] - nfft, step)
    for i in win_starts:
        sub = x[i : i + n]
        res.append(np.fft.fft(sub * win))
    return np.array(res, "complex"), win_starts


def make_specgram(
    sound, rate, shift_fft=True, sample_times=None, nfft=None, noverlap=2000, par=0
):
    assert rate > 1  # probably should have multiple samples per second
    if nfft is None:
        nfft = int(np.ceil(0.05 * rate))

    nfft += nfft % 2

    win = np.hamming(nfft)
    sft, time_idx = stfft(sound, nfft, noverlap, win, par=par)
    sft = np.real(sft * np.conjugate(sft))
    sft /= np.sum(np.abs(win) ** 2)
    freqs = np.fft.fftfreq(sft.shape[1], 1.0 / rate)

    # Since the input is real, the result will be symmetric, and thus we can throw away
    # the negative frequencies.
    nfreq = nfft // 2
    assert (freqs[nfreq - 1] > 0) and (freqs[nfreq] < 0)
    freqs = freqs[nfreq - 1 : 0 : -1]
    sft = sft[:, nfreq - 1 : 0 : -1]

    if sample_times is None:
        times = time_idx * (1.0 / rate)
    else:
        times = sample_times[time_idx]

    return freqs, np.asarray(sft, dtype="float32"), times


def test_spectrogram():
    # http://matplotlib.org/examples/pylab_examples/specgram_demo.html
    dt = 1.0 / 0.0005
    t = np.arange(0.0, 20.0, dt)
    # t = np.arange(0., 3., dt)
    s1 = np.sin((2 * np.pi) * 100 * t)
    s2 = 2 * np.sin((2 * np.pi) * 400 * t)
    s2[-((10 < t) & (t < 12))] = 0
    nse = 0.01 * np.random.randn(len(t))
    if 0:
        x = s1
    else:
        x = s1 + s2 + nse
    freqs, spec, spec_times = make_specgram(x, dt)

    pl.clf()

    ax1 = pl.subplot(211)
    ax1.plot(t, x)

    if 1:
        lsp = spec.copy()
        lsp[spec > 0] = np.log(spec[spec > 0])
        lsp = ut.clip_rescale(lsp, -10, np.percentile(lsp, 99))
    else:
        lsp = spec.copy()
        lsp = ut.clip_rescale(lsp, 0, np.percentile(lsp, 99))

    ax2 = pl.subplot(212, sharex=ax1)
    ax2.imshow(
        lsp.T,
        cmap=pl.cm.jet,
        extent=(0.0, t[-1], np.min(freqs), np.max(freqs)),
        aspect="auto",
    )

    ig.show(vis_specgram(freqs, spec, spec_times))
    ut.toplevel_locals()


def pink_noise(n, scale=1.0, alpha=1.0):
    # not exactly pink
    if n <= 1:
        return np.randn(n)
    spec = np.random.randn(n)
    # power \prop 1/f^alpha ==> sqrt(power) \prop 1/sqrt(f^alpha)
    spec = 1.0 / (scale * np.sqrt(np.arange(1, n + 1) ** alpha)) * spec
    return np.fft.irfft(spec)[:n]


class Sound:
    def __init__(self, times, rate, samples=None):
        # Allow Sound(samples, sr)
        if samples is None:
            samples = times
            times = None
        if samples.dtype == np.float32:
            samples = samples.astype("float64")

        self.rate = rate
        self.samples = ut.atleast_2d_col(samples)
        self.length = samples.shape[0]
        if times is None:
            self.times = np.arange(len(self.samples)) / float(self.rate)
        else:
            self.times = times

    def copy(self):
        return copy.deepcopy(self)

    def parts(self):
        return (self.times, self.rate, self.samples)

    def __getitem__(self, *args):
        return Sound(
            self.times.__getitem__(*args), self.rate, self.samples.__getitem__(*args)
        )

    def duration(self):
        return self.samples.shape[0] / float(self.rate)

    def normalized(self, check=True):
        if self.samples.dtype == np.double:
            assert (not check) or np.max(np.abs(self.samples)) <= 4.0
            x = copy.deepcopy(self)
            x.samples = np.clip(x.samples, -1.0, 1.0)
            return x
        else:
            s = copy.deepcopy(self)
            s.samples = np.array(s.samples, "double") / np.iinfo(s.samples.dtype).max
            s.samples[s.samples < -1] = -1
            s.samples[s.samples > 1] = 1
            return s

    def unnormalized(self, dtype_name="int32"):
        s = self.normalized()
        inf = np.iinfo(np.dtype(dtype_name))
        samples = np.clip(s.samples, -1.0, 1.0)
        samples = inf.max * samples
        samples = np.array(np.clip(samples, inf.min, inf.max), dtype_name)
        s.samples = samples
        return s

    def sample_from_time(self, t, bound=False):
        if t is None:
            return None

        if bound:
            return min(max(0, int(np.round(t * self.rate))), self.samples.shape[0] - 1)
        else:
            return int(np.round(t * self.rate))
        # if self.times[0] != 0:
        #   return int(np.argmin(np.abs(self.times - t)))
        # else:
        #   return min(max(0, int(np.round(t * self.rate))), self.samples.shape[0]-1)

    st = sample_from_time

    def shift_zero(self):
        s = copy.deepcopy(self)
        s.times -= s.times[0]
        return s

    def select_channel(self, c):
        s = copy.deepcopy(self)
        s.samples = s.samples[:, c]
        return s

    def left_pad_silence(self, n):
        if n == 0:
            return self.shift_zero()
        else:
            if np.ndim(self.samples) == 1:
                samples = np.concatenate([[0] * n, self.samples])
            else:
                samples = np.vstack(
                    [
                        np.zeros((n, self.samples.shape[1]), self.samples.dtype),
                        self.samples,
                    ]
                )
        return Sound(None, self.rate, samples)

    def right_pad_silence(self, n):
        if n == 0:
            return self.shift_zero()
        else:
            if np.ndim(self.samples) == 1:
                samples = np.concatenate([self.samples, [0] * n])
            else:
                samples = np.vstack(
                    [
                        self.samples,
                        np.zeros((n, self.samples.shape[1]), self.samples.dtype),
                    ]
                )
        return Sound(None, self.rate, samples)

    def pad_slice(self, s1, s2):
        assert s1 < self.samples.shape[0] and s2 >= 0
        s = self[max(0, s1) : min(s2, self.samples.shape[0])]
        s = s.left_pad_silence(max(0, -s1))
        s = s.right_pad_silence(max(0, s2 - self.samples.shape[0]))
        return s

    def to_mono(self, force_copy=True):
        s = copy.deepcopy(self)
        s.samples = make_mono(s.samples)
        return s

    def slice_time(self, t1, t2=None):
        return self[self.st(t1) : self.st(t2)]

    @property
    def nchannels(self):
        return 1 if np.ndim(self.samples) == 1 else self.samples.shape[1]

    def save(self, fname):
        s = self.unnormalized("int16")
        scipy.io.wavfile.write(fname, s.rate, s.samples)

    def resampled(self, new_rate, clip=True):
        from neurovc.contrib.flowmag.flow_models.pwcnet.aolib import subband

        if new_rate == self.rate:
            return copy.deepcopy(self)
        else:
            # assert self.samples.shape[1] == 1
            return Sound(
                None,
                new_rate,
                subband.resample(self.samples, float(new_rate) / self.rate, clip=clip),
            )

    def trim_to_size(self, n):
        return Sound(None, self.rate, self.samples[:n])


def play(samples_or_snd, sr=None, compress=True):
    if sr is None:
        samples = samples_or_snd.samples
        sr = samples_or_snd.rate
    else:
        samples = samples_or_snd
    snd = Sound(None, sr, samples)  # .unnormalized('int16')

    path = ut.pjoin(imtable.get_www_path(), "sounds")
    if compress:
        with ut.temp_file(".wav") as wav_fname:
            fname = ut.make_temp(".mp3", dir=path)
            # scipy.io.wavfile.write(wav_fname, snd.rate, snd.samples)
            snd.save(wav_fname)
            os.system('ffmpeg -loglevel error -y -i "%s" "%s"' % (wav_fname, fname))
    else:
        fname = ut.make_temp(".wav", dir=path)
        scipy.io.wavfile.write(fname, snd.rate, snd.samples)

    os.system("chmod a+rwx %s" % fname)
    # url = ut.pjoin(imtable.PUBLIC_URL, 'sounds', os.path.split(fname)[1])
    url = ut.pjoin(imtable.get_url(), "sounds", os.path.split(fname)[1])
    print(url)
    return url


class LongWav:
    def __init__(self, fname):
        self.wav = wave.open(fname)
        self.rate = int(self.wav.getframerate())
        # 16 bit
        if self.wav.getsampwidth() != 2:
            raise RuntimeError("Expected 16-bit wave file!")
        self.length = self.wav.getnframes()

    def __getitem__(self, i, j):
        self.wav.setpos(i)
        # print i, j
        data = self.wav.readframes(j - i)
        data = np.fromstring(data, dtype=np.int16)
        return Sound(None, self.rate, data)

    def sound(self):
        return self[: self.length]

    def duration(self):
        return self.length / float(self.rate)

    def sample_from_time(self, t, bound=False):
        if bound:
            return int(min(max(0, int(np.round(t * self.rate))), self.length - 1))
        else:
            return int(np.round(t * self.rate))

    def slice_time(self, t1, t2):
        return self[self.st(t1) : self.st(t2)]

    def pad_slice(self, s1, s2):
        assert s1 < self.length and s2 >= 0
        s = self[max(0, s1) : min(s2, self.length)]
        s = s.left_pad_silence(max(0, -s1))
        s = s.right_pad_silence(max(0, s2 - self.length))
        return s

    st = sample_from_time


def resample_snd(xxx_todo_changeme):
    (snd, sr) = xxx_todo_changeme
    return snd.resampled(sr)


def concat_sounds(snds, par=1):
    # snds = [snd.to_mono() for snd in snds]
    if ut.ndistinct(ut.mapattr(snds).rate) == 1:
        return Sound(
            None, snds[0].rate, np.concatenate(ut.mapattr(snds).samples, axis=0)
        )
    else:
        sr = max(ut.mapattr(snds).rate)
        # new_snds = [snd.resampled(sr) for snd in snds]

        if par:
            new_snds = ut.parmap(resample_snd, [(snd, sr) for snd in snds])
        else:
            new_snds = list(map(resample_snd, [(snd, sr) for snd in snds]))

        assert ut.ndistinct(ut.mapattr(new_snds).rate) == 1
        return concat_sounds(new_snds)


def convert_sound_compat(
    in_fname, out_fname, duration=None, rate=None, codec="pcm_s32le", ffmpeg_flags=""
):
    """Probably there is a better way to do this..."""
    duration_flag = "-t %.7f" % duration if duration is not None else ""
    assert out_fname.endswith(".wav")
    # with ut.temp_file('.mp3') as tmp_file:
    #   print tmp_file
    #   ok = (0 == ut.sys_print('ffmpeg -i "%s" %s -y "%s"' % (in_fname, duration_flag, tmp_file)))
    #   ok = ok and (0 == os.system('ffmpeg -i "%s" -y -ac 1 -acodec pcm_s32le "%s"' % (tmp_file, out_fname)))
    #   print 'writing', out_fname, ok
    #   #ok = ok and (0 == ut.sys_print('ffmpeg -i "%s" -y -f s16le -acodec pcm_s16le -ac 1 "%s"' % (tmp_file, out_fname)))
    # asdf
    # with ut.temp_file('.mp3') as mp3_tmp, ut.temp_file('.wav') as wav_tmp:
    with ut.temp_file(".mp3") as mp3_tmp, ut.temp_file(".wav") as wav_tmp:
        ok = 0 == ut.sys_print(
            'ffmpeg -i "%s" %s -y %s "%s"'
            % (in_fname, duration_flag, ffmpeg_flags, mp3_tmp)
        )
        # ok = ok and (0 == os.system('ffmpeg -i "%s" -y -ac 1 -acodec pcm_s32le "%s"' % (mp3_tmp, wav_tmp)))
        ar_str = "-ar %d" % rate if rate is not None else ""
        ok = ok and (
            0
            == os.system(
                'ffmpeg -i "%s" -y -ac 1 -acodec %s %s "%s"'
                % (mp3_tmp, codec, ar_str, wav_tmp)
            )
        )
        if ok:
            load_sound(wav_tmp).save(out_fname)
        print("writing", out_fname, ok)
        # ok = ok and (0 == ut.sys_print('ffmpeg -i "%s" -y -f s16le -acodec pcm_s16le -ac 1 "%s"' % (tmp_file, out_fname)))
    return ok


# def convert_sound_compat(in_fname, out_fname, duration = None, rate = None, codec = 'pcm_s32le', ffmpeg_flags = ''):
#   """ Probably there is a better way to do this... """
#   duration_flag = ('-t %.7f' % duration if duration is not None else '')
#   assert out_fname.endswith('.wav')
#   # with ut.temp_file('.mp3') as tmp_file:
#   #   print tmp_file
#   #   ok = (0 == ut.sys_print('ffmpeg -i "%s" %s -y "%s"' % (in_fname, duration_flag, tmp_file)))
#   #   ok = ok and (0 == os.system('ffmpeg -i "%s" -y -ac 1 -acodec pcm_s32le "%s"' % (tmp_file, out_fname)))
#   #   print 'writing', out_fname, ok
#   #   #ok = ok and (0 == ut.sys_print('ffmpeg -i "%s" -y -f s16le -acodec pcm_s16le -ac 1 "%s"' % (tmp_file, out_fname)))
#   #asdf
#   #with ut.temp_file('.mp3') as mp3_tmp, ut.temp_file('.wav') as wav_tmp:
#   with ut.temp_file('.mp3') as mp3_tmp, ut.temp_file('.wav') as wav_tmp:
#     ok = (0 == ut.sys_print('ffmpeg -i "%s" %s -y %s "%s"' % (in_fname, duration_flag, ffmpeg_flags, mp3_tmp)))
#     #ok = ok and (0 == os.system('ffmpeg -i "%s" -y -ac 1 -acodec pcm_s32le "%s"' % (mp3_tmp, wav_tmp)))
#     ar_str = '-ar %d' % rate if rate is not None else ''
#     ok = ok and (0 == os.system('ffmpeg -i "%s" -y -ac 1 -acodec %s %s "%s"' % (mp3_tmp, codec, ar_str, out_fname)))
#     # if ok:
#     #   load_sound(wav_tmp).save(out_fname)
#     print 'writing', out_fname, ok
#     #ok = ok and (0 == ut.sys_print('ffmpeg -i "%s" -y -f s16le -acodec pcm_s16le -ac 1 "%s"' % (tmp_file, out_fname)))
#   return ok


def audio_sampling_rate(fname):
    return int(
        ut.sys_with_stdout("ffprobe -show_streams %s | grep sample_rate" % fname).split(
            "sample_rate="
        )[1]
    )



try:
    from neurovc.contrib.raft.raft_helpers import RAFTOpticalFlow as _RAFTOpticalFlow

    HAS_TORCH = True
except Exception:  # noqa: BLE001
    HAS_TORCH = False

    class RAFTOpticalFlow:  # type: ignore[no-redef]
        def __init__(self, *_, **__):
            raise ImportError(
                "torch and related dependencies are required for RAFTOpticalFlow. "
                "Install the torch extra: 'pip install neurovc[torch]'"
            )
else:
    RAFTOpticalFlow = _RAFTOpticalFlow  # type: ignore[assignment]
    del _RAFTOpticalFlow


__all__ = [
    "HAS_TORCH",
    "RAFTOpticalFlow",
]



import torch.nn as nn
from einops import rearrange, repeat

from neurovc.contrib.flowmag.flow_utils import normalize_flow, warp


class MMLoss(nn.Module):
    def __init__(self, config):
        super().__init__()

        # Choose RAFT to calculate flow
        if config.train.flow_model == "raft":
            from neurovc.contrib.flowmag.flow_utils import RAFT

            self.flow_net = RAFT(num_iters=config.train.raft_iters)

        # Choose ARFlow to calculate flow
        elif config.train.flow_model == "arflow":
            from neurovc.contrib.flowmag.flow_utils import ARFlow

            self.flow_net = ARFlow()

        else:
            raise NotImplementedError

        # Use L1 for loss calculation
        self.criterion = nn.L1Loss()

    def compute_flow(self, im, frames, backward=False):
        """
        Given an image (b, c, h, w) and frames (b, c, t, h, w) we compute
            the flow between the image and all frames and return them
            in the shape (b, t, 2, h, w)
        If backward == True, then reverse the direction: compute flow from
            all frames to single image
        """
        b, c, t, h, w = frames.shape

        # Repeat and rearrange as needed
        im = repeat(im, "b c h w -> (b num_frames) c h w", num_frames=t)
        frames = rearrange(frames, "b c t h w -> (b t) c h w")

        # Compute flow
        if backward:
            flow = self.flow_net(frames, im)
        else:
            flow = self.flow_net(im, frames)

        # Rearrange to separate batch and time dim
        flow = rearrange(flow, "(b t) d h w -> b t d h w", b=b, t=t)

        return flow

    def compute_mag_loss(self, flow_pred, flow_src, alphas):
        # Enforce predicted flow == alpha * src flow
        return self.criterion(flow_pred, alphas * flow_src)

    def warp_frames(self, frames, flow):
        # Combine batch and time dims
        flow = rearrange(flow, "b t d h w -> (b t) d h w")

        # Normalize flow
        flow = normalize_flow(flow)

        # Rearrange frames to match flows
        frames = rearrange(frames, "b c t h w -> (b t) c h w")

        # Warp future frames into first
        frames_warped = warp(frames, flow)

        return frames_warped

    def compute_color_loss(self, pred, src, flow_pred, flow_src):
        info = {}

        # Warp frames into first
        src_warped = self.warp_frames(src[:, :, 1:], flow_src)
        pred_warped = self.warp_frames(pred, flow_pred)

        # Enforce the pixels in pred videos keep consistent with the input video
        color_loss = self.criterion(pred_warped, src_warped)

        # Add warps to info
        info["pred_warped"] = pred_warped
        info["src_warped"] = src_warped

        return color_loss, info

    def forward(self, pred, src, alphas):
        """
        pred: magnified future frame predictions (b, c, t-1, h, w)
        src: original frames (b, c, t, h, w)
        alphas: magnification factors (per batch)
        """

        # Add dims to alpha to match flow (b, t, d, h, w)
        if len(alphas.shape) == 1:
            # Only batch dim, constant alpha per video
            alphas = alphas.view(-1, 1, 1, 1, 1)
        elif len(alphas.shape) == 3:
            # Spatially varying alpha, shape (b, h, w)
            b, h, w = alphas.shape
            alphas = alphas.view(b, 1, 1, h, w)

        # Get flows
        anchor_frame = src[:, :, 0]
        flow_pred = self.compute_flow(anchor_frame, pred)
        flow_src = self.compute_flow(anchor_frame, src[:, :, 1:])

        # Compute losses
        mag_loss = self.compute_mag_loss(flow_pred, flow_src, alphas)
        color_loss, info = self.compute_color_loss(pred, src, flow_pred, flow_src)

        # Add to info
        info["flow_pred"] = flow_pred
        info["flow_src"] = flow_src

        return mag_loss, color_loss, info



import json
import pprint
import datetime
import argparse
from path import Path
from easydict import EasyDict

import basic_train
from logger import init_logger


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", default="configs/sintel_ft.json")
    parser.add_argument("-e", "--evaluate", action="store_true")
    parser.add_argument("-m", "--model", default=None)
    parser.add_argument("--n_gpu", type=int, default=1)
    args = parser.parse_args()

    with open(args.config) as f:
        cfg = EasyDict(json.load(f))

    if args.evaluate:
        cfg.train.update(
            {
                "epochs": 1,
                "epoch_size": -1,
                "valid_size": 0,
                "workers": 1,
                "val_epoch_size": 1,
            }
        )

    if args.model is not None:
        cfg.train.pretrained_model = args.model
    cfg.train.n_gpu = args.n_gpu

    # store files day by day
    curr_time = datetime.datetime.now().strftime("%y%m%d%H%M%S")
    cfg.save_root = Path("./outputs/checkpoints") / curr_time[:6] / curr_time[6:]
    cfg.save_root.makedirs_p()

    # init logger
    _log = init_logger(log_dir=cfg.save_root, filename=curr_time[6:] + ".log")
    _log.info("=> will save everything to {}".format(cfg.save_root))

    # show configurations
    cfg_str = pprint.pformat(cfg)
    _log.info("=> configurations \n " + cfg_str)

    basic_train.main(cfg, _log)



# Part of the code from https://github.com/visinf/irr/blob/master/augmentations.py

import torch
import torch.nn as nn
from transforms.ar_transforms.interpolation import Interp2
from transforms.ar_transforms.interpolation import Meshgrid
import numpy as np


def denormalize_coords(xx, yy, width, height):
    """scale indices from [-1, 1] to [0, width/height]"""
    xx = 0.5 * (width - 1.0) * (xx.float() + 1.0)
    yy = 0.5 * (height - 1.0) * (yy.float() + 1.0)
    return xx, yy


def normalize_coords(xx, yy, width, height):
    """scale indices from [0, width/height] to [-1, 1]"""
    xx = (2.0 / (width - 1.0)) * xx.float() - 1.0
    yy = (2.0 / (height - 1.0)) * yy.float() - 1.0
    return xx, yy


def apply_transform_to_params(theta0, theta_transform):
    a1 = theta0[:, 0]
    a2 = theta0[:, 1]
    a3 = theta0[:, 2]
    a4 = theta0[:, 3]
    a5 = theta0[:, 4]
    a6 = theta0[:, 5]
    #
    b1 = theta_transform[:, 0]
    b2 = theta_transform[:, 1]
    b3 = theta_transform[:, 2]
    b4 = theta_transform[:, 3]
    b5 = theta_transform[:, 4]
    b6 = theta_transform[:, 5]
    #
    c1 = a1 * b1 + a4 * b2
    c2 = a2 * b1 + a5 * b2
    c3 = b3 + a3 * b1 + a6 * b2
    c4 = a1 * b4 + a4 * b5
    c5 = a2 * b4 + a5 * b5
    c6 = b6 + a3 * b4 + a6 * b5
    #
    new_theta = torch.stack([c1, c2, c3, c4, c5, c6], dim=1)
    return new_theta


class _IdentityParams(nn.Module):
    def __init__(self):
        super(_IdentityParams, self).__init__()
        self._batch_size = 0
        self.register_buffer("_o", torch.FloatTensor())
        self.register_buffer("_i", torch.FloatTensor())

    def _update(self, batch_size):
        torch.zeros([batch_size, 1], out=self._o)
        torch.ones([batch_size, 1], out=self._i)
        return torch.cat([self._i, self._o, self._o, self._o, self._i, self._o], dim=1)

    def forward(self, batch_size):
        if self._batch_size != batch_size:
            self._identity_params = self._update(batch_size)
            self._batch_size = batch_size
        return self._identity_params


class RandomMirror(nn.Module):
    def __init__(self, vertical=True, p=0.5):
        super(RandomMirror, self).__init__()
        self._batch_size = 0
        self._p = p
        self._vertical = vertical
        self.register_buffer("_mirror_probs", torch.FloatTensor())

    def update_probs(self, batch_size):
        torch.ones([batch_size, 1], out=self._mirror_probs)
        self._mirror_probs *= self._p

    def forward(self, theta_list):
        batch_size = theta_list[0].size(0)
        if batch_size != self._batch_size:
            self.update_probs(batch_size)
            self._batch_size = batch_size

        # apply random sign to a1 a2 a3 (these are the guys responsible for x)
        sign = torch.sign(2.0 * torch.bernoulli(self._mirror_probs) - 1.0)
        i = torch.ones_like(sign)
        horizontal_mirror = torch.cat([sign, sign, sign, i, i, i], dim=1)
        theta_list = [theta * horizontal_mirror for theta in theta_list]

        # apply random sign to a4 a5 a6 (these are the guys responsible for y)
        if self._vertical:
            sign = torch.sign(2.0 * torch.bernoulli(self._mirror_probs) - 1.0)
            vertical_mirror = torch.cat([i, i, i, sign, sign, sign], dim=1)
            theta_list = [theta * vertical_mirror for theta in theta_list]

        return theta_list


class RandomAffineFlow(nn.Module):
    def __init__(self, cfg, addnoise=True):
        super(RandomAffineFlow, self).__init__()
        self.cfg = cfg
        self._interp2 = Interp2(clamp=False)
        self._flow_interp2 = Interp2(clamp=False)
        self._meshgrid = Meshgrid()
        self._identity = _IdentityParams()
        self._random_mirror = (
            RandomMirror(cfg.vflip) if cfg.hflip else RandomMirror(p=1)
        )
        self._addnoise = addnoise

        self.register_buffer("_noise1", torch.FloatTensor())
        self.register_buffer("_noise2", torch.FloatTensor())
        self.register_buffer("_xbounds", torch.FloatTensor([-1, -1, 1, 1]))
        self.register_buffer("_ybounds", torch.FloatTensor([-1, 1, -1, 1]))
        self.register_buffer("_x", torch.IntTensor(1))
        self.register_buffer("_y", torch.IntTensor(1))

    def inverse_transform_coords(
        self, width, height, thetas, offset_x=None, offset_y=None
    ):
        xx, yy = self._meshgrid(width=width, height=height)

        xx = torch.unsqueeze(xx, dim=0).float()
        yy = torch.unsqueeze(yy, dim=0).float()

        if offset_x is not None:
            xx = xx + offset_x
        if offset_y is not None:
            yy = yy + offset_y

        a1 = thetas[:, 0].contiguous().view(-1, 1, 1)
        a2 = thetas[:, 1].contiguous().view(-1, 1, 1)
        a3 = thetas[:, 2].contiguous().view(-1, 1, 1)
        a4 = thetas[:, 3].contiguous().view(-1, 1, 1)
        a5 = thetas[:, 4].contiguous().view(-1, 1, 1)
        a6 = thetas[:, 5].contiguous().view(-1, 1, 1)

        xx, yy = normalize_coords(xx, yy, width=width, height=height)
        xq = a1 * xx + a2 * yy + a3
        yq = a4 * xx + a5 * yy + a6
        xq, yq = denormalize_coords(xq, yq, width=width, height=height)
        return xq, yq

    def transform_coords(self, width, height, thetas):
        xx1, yy1 = self._meshgrid(width=width, height=height)
        xx, yy = normalize_coords(xx1, yy1, width=width, height=height)

        def _unsqueeze12(u):
            return torch.unsqueeze(torch.unsqueeze(u, dim=1), dim=1)

        a1 = _unsqueeze12(thetas[:, 0])
        a2 = _unsqueeze12(thetas[:, 1])
        a3 = _unsqueeze12(thetas[:, 2])
        a4 = _unsqueeze12(thetas[:, 3])
        a5 = _unsqueeze12(thetas[:, 4])
        a6 = _unsqueeze12(thetas[:, 5])
        #
        z = a1 * a5 - a2 * a4
        b1 = a5 / z
        b2 = -a2 / z
        b4 = -a4 / z
        b5 = a1 / z
        #
        xhat = xx - a3
        yhat = yy - a6
        xq = b1 * xhat + b2 * yhat
        yq = b4 * xhat + b5 * yhat

        xq, yq = denormalize_coords(xq, yq, width=width, height=height)
        return xq, yq

    def find_invalid(self, width, height, thetas):
        x = self._xbounds
        y = self._ybounds
        #
        a1 = torch.unsqueeze(thetas[:, 0], dim=1)
        a2 = torch.unsqueeze(thetas[:, 1], dim=1)
        a3 = torch.unsqueeze(thetas[:, 2], dim=1)
        a4 = torch.unsqueeze(thetas[:, 3], dim=1)
        a5 = torch.unsqueeze(thetas[:, 4], dim=1)
        a6 = torch.unsqueeze(thetas[:, 5], dim=1)
        #
        z = a1 * a5 - a2 * a4
        b1 = a5 / z
        b2 = -a2 / z
        b4 = -a4 / z
        b5 = a1 / z
        #
        xhat = x - a3
        yhat = y - a6
        xq = b1 * xhat + b2 * yhat
        yq = b4 * xhat + b5 * yhat
        xq, yq = denormalize_coords(xq, yq, width=width, height=height)
        #
        invalid = ((xq < 0) | (yq < 0) | (xq >= width) | (yq >= height)).sum(
            dim=1, keepdim=True
        ) > 0

        return invalid

    def apply_random_transforms_to_params(
        self,
        theta0,
        max_translate,
        min_zoom,
        max_zoom,
        min_squeeze,
        max_squeeze,
        min_rotate,
        max_rotate,
        validate_size=None,
    ):
        max_translate *= 0.5
        batch_size = theta0.size(0)
        height, width = validate_size

        # collect valid params here
        thetas = torch.zeros_like(theta0)

        zoom = theta0.new(batch_size, 1).zero_()
        squeeze = torch.zeros_like(zoom)
        tx = torch.zeros_like(zoom)
        ty = torch.zeros_like(zoom)
        phi = torch.zeros_like(zoom)
        invalid = torch.ones_like(zoom).byte()

        while invalid.sum() > 0:
            # random sampling
            zoom.uniform_(min_zoom, max_zoom)
            squeeze.uniform_(min_squeeze, max_squeeze)
            tx.uniform_(-max_translate, max_translate)
            ty.uniform_(-max_translate, max_translate)
            phi.uniform_(min_rotate, max_rotate)

            # construct affine parameters
            sx = zoom * squeeze
            sy = zoom / squeeze
            sin_phi = torch.sin(phi)
            cos_phi = torch.cos(phi)
            b1 = cos_phi * sx
            b2 = sin_phi * sy
            b3 = tx
            b4 = -sin_phi * sx
            b5 = cos_phi * sy
            b6 = ty

            theta_transform = torch.cat([b1, b2, b3, b4, b5, b6], dim=1)
            theta_try = apply_transform_to_params(theta0, theta_transform)
            thetas = invalid.float() * theta_try + (1 - invalid).float() * thetas

            # compute new invalid ones
            invalid = self.find_invalid(width=width, height=height, thetas=thetas)

        # here we should have good thetas within borders
        return thetas

    def transform_image(self, images, thetas):
        batch_size, channels, height, width = images.size()
        xq, yq = self.transform_coords(width=width, height=height, thetas=thetas)
        transformed = self._interp2(images, xq, yq)
        return transformed

    def transform_flow(self, flow, theta1, theta2):
        batch_size, channels, height, width = flow.size()
        u = flow[:, 0, :, :]
        v = flow[:, 1, :, :]

        # inverse transform coords
        x0, y0 = self.inverse_transform_coords(
            width=width, height=height, thetas=theta1
        )

        x1, y1 = self.inverse_transform_coords(
            width=width, height=height, thetas=theta2, offset_x=u, offset_y=v
        )

        # subtract and create new flow
        u = x1 - x0
        v = y1 - y0
        new_flow = torch.stack([u, v], dim=1)

        # transform coords
        xq, yq = self.transform_coords(width=width, height=height, thetas=theta1)

        # interp2
        transformed = self._flow_interp2(new_flow, xq, yq)
        return transformed

    def forward(self, data):
        # 01234 flow 12 21 23 32
        imgs = data["imgs"]
        flows_f = data["flows_f"]
        masks_f = data["masks_f"]

        batch_size, _, height, width = imgs[0].size()

        # identity = no transform
        theta0 = self._identity(batch_size)

        # global transform
        theta_list = [
            self.apply_random_transforms_to_params(
                theta0,
                max_translate=self.cfg.trans[0],
                min_zoom=self.cfg.zoom[0],
                max_zoom=self.cfg.zoom[1],
                min_squeeze=self.cfg.squeeze[0],
                max_squeeze=self.cfg.squeeze[1],
                min_rotate=self.cfg.rotate[0],
                max_rotate=self.cfg.rotate[1],
                validate_size=[height, width],
            )
        ]

        # relative transform
        for i in range(len(imgs) - 1):
            theta_list.append(
                self.apply_random_transforms_to_params(
                    theta_list[-1],
                    max_translate=self.cfg.trans[1],
                    min_zoom=self.cfg.zoom[2],
                    max_zoom=self.cfg.zoom[3],
                    min_squeeze=self.cfg.squeeze[2],
                    max_squeeze=self.cfg.squeeze[3],
                    min_rotate=self.cfg.rotate[2],
                    max_rotate=self.cfg.rotate[2],
                    validate_size=[height, width],
                )
            )

        # random flip images
        theta_list = self._random_mirror(theta_list)

        # 01234
        imgs = [self.transform_image(im, theta) for im, theta in zip(imgs, theta_list)]

        if len(imgs) > 2:
            theta_list = theta_list[1:-1]
        # 12 23
        flows_f = [
            self.transform_flow(flo, theta1, theta2)
            for flo, theta1, theta2 in zip(flows_f, theta_list[:-1], theta_list[1:])
        ]

        masks_f = [
            self.transform_image(mask, theta)
            for mask, theta in zip(masks_f, theta_list)
        ]

        if self._addnoise:
            stddev = np.random.uniform(0.0, 0.04)
            for im in imgs:
                noise = torch.zeros_like(im)
                noise.normal_(std=stddev)
                im.add_(noise)
                im.clamp_(0.0, 1.0)

        data["imgs"] = imgs
        data["flows_f"] = flows_f
        data["masks_f"] = masks_f
        return data
